import { ApolloServer } from 'apollo-server-fastify';
import fastify from 'fastify';
import { DocumentNode, print } from 'graphql';

export type StringOrAst = string | DocumentNode;
export type Options<T extends object> = { variables?: T };

export type TestClientConfig = {
  // The ApolloServer instance that will be used for handling the queries you run in your tests.
  // Must be an instance of the ApolloServer class from `apollo-server-fastify` (or a compatible subclass).
  apolloServer: ApolloServer;
  // Extends the mocked Request object with additional keys.
  // Useful when your apolloServer `context` option is a callback that operates on the passed in `req` key,
  // and you want to inject data into that `req` object.
  extendMockRequest?: fastify.HTTPInjectOptions;
};

export type TestQuery = <T extends object = {}, V extends object = {}>(
  operation: StringOrAst,
  { variables }?: Options<V>
) => Promise<T>;

export type TestSetOptions = (options: {
  request?: fastify.HTTPInjectOptions;
}) => void;

// This function takes in an apollo server instance and returns a function that you can use to run operations
// against your schema, and assert on the results.
//
// Example usage:
// ```
// const apolloServer = await createApolloServer({ schema })
// const query = createTestClient({
//   apolloServer,
// });
//
// const result = await query(
//   `{ currentUser { id } }`
// )
//
// expect(result).toEqual({
//   data: {
//     currentUser: {
//       id: '1'
//     }
//   }
// });
// ```
export function createTestClient({
  apolloServer,
  extendMockRequest = {},
}: TestClientConfig) {
  const app = fastify();
  app.register(apolloServer.createHandler());

  let mockRequestOptions = extendMockRequest;

  /**
   * Set the options after TestClient creation
   * Useful when you don't want to create a new instance just for a specific change in the request.
   */
  const setOptions: TestSetOptions = ({
    request,
  }: {
    request?: fastify.HTTPInjectOptions;
  }) => {
    if (request) {
      mockRequestOptions = request;
    }
  };

  const test: TestQuery = async <T extends object = {}, V extends object = {}>(
    operation: StringOrAst,
    { variables }: Options<V> = {}
  ) => {
    const injectOptions: fastify.HTTPInjectOptions = {
      url: apolloServer.graphqlPath,
      method: 'POST',
      payload: {
        query: (typeof operation === 'string' ? operation : print(operation)),
        variables,
      },
      ...mockRequestOptions,
    };

    const graphqlResponse = await app.inject(injectOptions).then((res) => res.json());

    return graphqlResponse as T;
  };

  return {
    query: test,
    mutate: test,
    setOptions
  };
}
